processSchemaTypes =
   # Process all the elements in types (hopefully <schema> elements)
   # and merge. We may want to keep these as a list indexed by namespace/schema identifier.
   # Is this still true?
  
   # Originally, we just processed the first one.
   # types = processSchemaTypes(doc[["types"]][[1]], doc)


  # We are called with the doc[["types"]] argument
  # That should contain a schema.  The schema
  # can have elements within it (e.g. interop.wsdl) or
  # import statements which give other schema, e.g. eutils.wsdl
  # or both import statements and new type defintions (e.g. KEGG.wsdl).
  #

  # This can be called recursively for sub-schema, i.e. schema defined within the
  # <types><schema>...</schema></types>

  # This is (initially) called with the top-level <types> which should contain one
  # <schema> node.

function(node, doc, namespaceDefs = xmlNamespaceDefinitions(doc), createConverters = TRUE, verbose = FALSE)
{
  if(is.null(node))
      return(list())

 substGroups = getSubstitutionGroups(doc)
  
     # Loop over the children of this node and process each element.
 ans <- list()
 lapply(1:xmlSize(node),
                function(i) {
                   el = node[[i]]

                   if(inherits(el, c("XMLCommentNode", "XMLInternalCommentNode"))
                         || xmlName(el) %in% c("import", "include"))
                     return(NULL)

                   if(xmlName(el) == "annotation")
                     return(NULL)
                   
                   if(xmlName(el) == "schema") {

                         # pubmed from NCBI has a <schema><schema targetNamespace="">...
                      if(xmlSize(el) == 1 && names(el) == "schema")
                        el = el[[1]]
                     
                      ns = xmlGetAttr(el, "targetNamespace")
                      if(verbose)
                         cat("processing (sub) schema", xmlGetAttr(el, "targetNamespace"), "\n")
                      o = processSchemaTypes(el, doc, namespaceDefs, createConverters = createConverters, verbose = verbose)
                      o = new("SchemaTypes", o)
                      if(length(o))
                        ans[[ns]] <<-  o
                   } else {
                      n = xmlGetAttr(el, "name", as.character(NA))
                      if(verbose)
                         cat(i,")", n, "\n")
                      o = processSchemaType(el, substitutionGroups = substGroups)
                      if(createConverters && is(o, "BasicSOAPType"))
                         o@fromConverter = createSOAPConverter(o)
                      ans[[n]] <<- o
                    }

                   NULL
                 })

         # Fix the names on these types to avoid the schema.
 #             types = unlist(types, recursive = FALSE)
 #names(types) = sapply(types, function(x) x$name)
 # types

  if(all(sapply(ans, inherits, "SchemaTypes"))) {
     ans = new("SchemaCollection", ans)
   }

#Debugging!
#cat(xmlName(node), xmlSize(node), "=>", class(ans), paste(sapply(ans, class), sep = ", "),"\n")
  ans
}




# This material relates to XML schema and processing of types defined within
# schema. Ideally, we will separate this into a separate package and use the
# results in different ways, not just for SOAP. And we will provide a more
# extensive framework for dealing with all of the schema details rather than
# the rather limited but common ones supported here.
#
#  See the O'Reilly book XML Schemas by Eric van der Vlist for more information
#  on schemas, or look on the Web (e.g. www.xml.com/schemas)
#
#  Priscilla Walmsley's Definitive XML Schema Prentice Hall
#  is much better. 
#
#
#

# Failed  complexType & attribute.
#   Assay_molTypeType
#   Assembly_currentType


processSchemaType =
  #
  # This is intended to create an R description for
  # XML Schema nodes that actually define a type.
  #
  #
  # This currently deals with a very small subset of XML schema
  # specifications.  These are the common ones, but more work 
  # is definitely needed.
  #

  #  simpleType
  #  complexType
  #  complexContent

   # sequence

  # Doesn't look like we use types.

  # Fix for
  #   simpleType with "attribute" child.
  #   simpleType with restriction.
function(type, types, substitutionGroups = NULL, namespaceDefs = list())
{

   if(inherits(type, c("XMLInternalCommentNode", "XMLComment")))
      return(NULL)
  
# Rationalize this 
  name = xmlGetAttr(type, "name", "")


  if(xmlSize(type) > 0 && "annotation" == xmlName(type[[1]])) # drop the annotation sub-node.
    xmlChildren(type) = xmlChildren(type)[-1]


     #   simpleType only and return.

  if(xmlName(type) == "element" && !is.na(xmlGetAttr(type, "type", NA)) && xmlSize(type) == 0) {
       # an element that really just refers to another type definition.
    refName = discardNamespace( xmlGetAttr(type, "type") ) #XXX stripping away name space but have to be careful if there is more than one.
    refNode = getNodeSet(as(type, "XMLInternalDocument"), paste("//x:*[@name=", sQuote(refName), "]"),
                            c("x" = "http://www.w3.org/2001/XMLSchema"))
    ref = if(length(refNode) == 0) 
             NULL
          else
             processSchemaType(refNode[[1]])
    return(new("Element", name = name, type = ref)) #XXX  attributes = ref@attributes,
    
  } else if(name == "simpleType" || xmlName(type) == "simpleType") {
    if(xmlSize(type) > 0 && xmlName(type[[1]]) == "restriction") {
       if(xmlSize(type[[1]]) > 0) {
            # check if the base type is a primitive (e.g. a double, ...)
            # and take care of the namespace, e.g. xs:double
         base = asQName(xmlGetAttr(type[[1]], "base"))
         if(!is.na(getRTypeFromSOAP(base[2], asIndex = TRUE))) { #XXX what if no prefix? Make asQName() return a vector of length 2?
#XXXX is this the right thing to do here at all?
           if(base[2] == "string") {
              def <- new("EnumValuesDef", name = name, values = xmlSApply(type[[1]],  xmlGetAttr, "value"))
           } else {
              tp = SOAPType(base[2], base[1], count = getElementCount(type))
              def = new("ExtendedClassDefinition", name = xmlGetAttr(type, "name", as.character(NA)), base = base[2])
           }
        } else
              # Assume enumeration values for present.
           def <- new("EnumValuesDef", name = name, values = xmlSApply(type[[1]],  xmlGetAttr, "value"))
       } else {
         # e.g. from eBaySvc.wsdl
         # <xs:simpleType name="DisputeIDType">
         #   <xs:restriction base="xs:string"/>
         # </xs:simpleType>
         def = new("ExtendedClassDefinition", name = name, base = xmlGetAttr(type[[1]], "base"))
       }
   } else if(xmlSize(type) > 0 && xmlName(type[[1]]) == "list") {
       if(xmlSize(type[[1]]) > 0) {
          el = processSchemaType(type[[1]][[1]], types, namespaceDefs)
          def = SOAPType(name, count = getElementCount(type), obj = new("RestrictedListType"))
          def@elType = el
          if(is(el, "EnumValuesDef"))  # And is a string
              def@elements = el@values
       } else if(xmlName(type) == "simpleType" && xmlName(type[[1]]) == "list"){
           def = processSimpleList(type[[1]],  xmlGetAttr(type, "name"))
       } else
           stop("Not sure what to do here with ", xmlName(type))
   } else
       def <- "xsd:string"

    return(def)
  }

   if("complexType" %in% names(type) && xmlSize(type[["complexType"]])  == 0) {
     return(new("Element", name = name, attributes = lapply(xmlChildren(type)[names(type) == "attribute"], processAttribute)))
   } else if(names(type)[1] == "complexType" && names(type[[1]]) == "simpleContent") {
     attrs = xmlApply(type[[1]][[1]][[1]], processAttribute)
     return(new("SimpleElement", name = xmlGetAttr(type, "name", as.character(NA)),
                                 attributes = attrs,
                                 type = xmlGetAttr(type[[1]][[1]][[1]], "base", character()), #XXX type should be the extension type
                                 count = getElementCount(type)))
  } else if(xmlName(type) %in% c("complexContent", "element"))
    tmp = type
  else if(xmlName(type) == "complexType" && xmlSize(type) == 0) {
          #??? What do we do here.  See "http://www.ebi.ac.uk/ebisearch/service.ebi?wsdl" for example.
    return(new("AnySOAPType", name = if(name != "") name else "AnySOAPType"))
  } else if(xmlName(type) == "complexType" && xmlName(type[[1]]) == "all")
    tmp = type
  else if(xmlName(type) == "complexType" && ( (xmlSize(type) == 1 && names(type) == "sequence")
                                              || ("sequence" %in% names(type)  && all(names(type) %in% c("attribute", "annotation", "sequence")))))  {
    return(processSequence(type[["sequence"]], types, namespaceDefs, name))
  } else if(xmlName(type) == 'complexType' && (xmlSize(type) == 1 && names(type) == "attributeGroup")) {
       # Just an attribute group within the complexType.

        #XXX We take off any names space. We should be more careful here if there is more than one namespace.
     groupName = discardNamespace( xmlGetAttr(type[["attributeGroup"]], "ref") )
     agroup = getNodeSet(as(type, "XMLInternalDocument"), paste("//x:attributeGroup[@name=", sQuote(groupName), "]"), "x")
     if(length(agroup) == 0)
       stop("Cannot find attribute group named ", sQuote(groupName))
     attrs = processAttributeGroup(agroup[[1]])
     return(new("SOAPComplexType", name = xmlGetAttr(type, "name"), attributes = attrs))
     
  } else if(xmlName(type) == "attribute") {
       # Case where xmlSize() > 0. But this must be (?)
       # a simpleType with a restriction on the type and so would be
       # handled in processAttribute(). So this could be simplified.
       # 
       # XXX is this right ? Think this must be handled by processAttribute() correctly.
       # We may have a restriction on the value. e.g. from the pmml schema
# <xs:attribute name="type" use="required">
#  <xs:simpleType>
#    <xs:restriction base="xs:string">
#      <xs:enumeration value="int"/>
#      <xs:enumeration value="real"/>
#      <xs:enumeration value="string"/>
#    </xs:restriction>
#  </xs:simpleType>
# </xs:attribute>
     return(processAttribute(type, name))
  } else 
    tmp <- type[["complexContent"]]

  if(xmlName(type) == "attributeGroup") {
        # We'll deal with these when they are referenced.
        # We could compute them just once and access that but where do we put them so that we can access them.
        # An environment like RGCCTranslationUnit.
      return(NULL)
      
  } else if(xmlName(type) == "element") {
       return(processSchemaElement(tmp))
  } else if(!is.null(tmp) && !is.null(tmp[["all"]])) {
        # Connect this with the other case !is.null(tmp <- type[["all"]]) below.
    
             # a struct-like definition with slots.
     a = tmp[["all"]]

     slotTypes = xmlApply(a, function(x) {
                                   # just return the SOAP type. We'll resolve it later.
                                  tt = xmlGetAttr(x, "type")
                                  if(length(tt) == 0) {
                                     tt = xmlGetAttr(x, "ref")
                                  }
                                  SOAPType(tt, nsuri = xmlNamespace(x), namespaceDefs = namespaceDefs, count = getElementCount(a))
                                    #getArrayElementTypeFromName
                               })

     names(slotTypes) = xmlSApply(a, getElementName)
     def = ClassDef(name, slotTypes)

  } else if(!is.null(tmp) && !is.null(tmp[["extension"]])) {

     base = xmlGetAttr(tmp[["extension"]], "base")
                      # Now get the extensions.
     def = ClassDef(name, list(), new("ExtendedClassDefinition"))
     def@base = base
     tmp = tmp[[1]]
     if(xmlSize(tmp) > 0 && xmlName(tmp[[1]]) == "sequence") {
               # ??? Should  we put this in a list as we now do?
       def@slotTypes = list(processSequence(tmp[[1]], types, namespaceDefs)) #XXX what about the name argument
     }

  } else if(!is.null(tmp)) {
    
                   # Currently we interpret this as an Array.
                   # So get the <restriction><attribute> element
                   # and take the arrayType and ref attribution from that.
            # In interop.wsdl, we also have a <sequence> <element> ... node within this restriction.
    restriction = tmp[["restriction"]]
    a = restriction[["attribute"]]
    if(is.null(a)) {
          # See pmml-3-2.xsd and the "row" element.
         #XXXXXXXX
        base =  xmlGetAttr(restriction, "base")
        if(xmlSize(restriction) > 0) {
             
        }
        def = NULL
    } else {
       def = ArrayType(xmlGetAttr(a, "arrayType", addNamespace = FALSE), ns = xmlNamespace(a), namespaceDefs = namespaceDefs)
       def@name = name
    }
    
  } else if(!is.null(tmp <- type[["all"]])) {
    
                    # a struct-like definition with slots.
     slotTypes = xmlApply(tmp, xmlGetAttr, "type")
     names(slotTypes) = xmlSApply(tmp, xmlGetAttr, "name")
     def = ClassDef(name, slotTypes, new("ArrayClassDefinition"))
     
  } else if(!is.null(tmp <- type[["sequence"]])) {
    
                    # <complexType><sequence>
                    # Connect this with the code for the "all" case.
                    # Need to resolve the type if it is not a primitive.
                    # XXX also want the minOccurs and maxOccurs
      def = SOAPType(name, counts = getElementCount(type), obj = new("SOAPComplexType"))
      def@content = processSequence(tmp, types, namespaceDefs)
      def@attributes = lapply(xmlChildren(type)[names(type) == "attribute"], processAttribute)
#      def = new("SimpleSequenceType", name = name, elType = slotTypes, elementType = slotTypes@name) 
#      def = ClassDef(name, slotTypes)
#      def@count = getElementCount(type)
      
  } else if(!is.null(tmp <- type[["choice"]])) {

      #XXX type or ref?
      # Handle local definitions
      slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type", xmlGetAttr(x, "ref", as.character(NA))))
      names(slotTypes) = xmlSApply(tmp,  xmlGetAttr, "name")
      if(xmlSize(tmp) == 1)
        def = ClassDef(name, slotTypes[1])
      else        #XXX Define UnionDef
        def = UnionDef(name, slotTypes)

   } else if(!is.null(tmp <- type[["simpleContent"]])) {
      if(xmlName(tmp[[1]]) == "extension") {
        
                      # Assume the children are attributes for the moment.
         attrTypes = xmlApply(tmp[[1]], processSchemaType)
         names(attrTypes) = xmlSApply(tmp[[1]],  xmlGetAttr, "name")
                      #XXX Add base.
         base = xmlGetAttr(tmp[[1]], "base")         
         def <- ClassDef(name, attrTypes)

      } else
          def <- NULL
  } else if(xmlName(type) == "complexType" &&
               xmlSize(type) > 0 && all(xmlSApply(type, xmlName) == "attribute")) {

        els = xmlApply(type, processAttribute)
        name = xmlGetAttr(type, "name")
        def = new("SOAPComplexType", name = name, attributes = els)

       # XXX
#<xs:complexType name="ArrayType" mixed="true">
# <xs:attribute name="n" type="INT-NUMBER" use="optional"/>
# <xs:attribute name="type" use="required">
#   <xs:simpleType>
#     <xs:restriction base="xs:string">
#       <xs:enumeration value="int"/>
#       <xs:enumeration value="real"/>
#       <xs:enumeration value="string"/>
#     </xs:restriction>
#   </xs:simpleType>
# </xs:attribute>
#</xs:complexType>

# <xs:complexType>
#  <xs:attribute name="name" type="xs:string" use="required"/>
#  <xs:attribute name="optype" type="OPTYPE"/>
#  <xs:attribute name="dataType" type="DATATYPE"/>
# </xs:complexType> 
    
     
  } else {
     warning("Failed to handle node ", name, " of type ", xmlName(type),
              if(xmlSize(type) > 0) c(" & ", xmlName(type[[1]])), " in processSchemaType. ",
                as(type, "character"),
                 class = "ProcessWSDLTypeError")
     return(NULL)
  }

  return(def)
 
  structure(list(name = name, definition = def),
            class = "WSDLTypeDescription")
}


getElementName =
  #
  # Handles an <element name="..."> and <element ref="...">
  #
function(node)
{
   ans = xmlGetAttr(x, "name")
   if(length(ans))
     return(ans)

   ref = xmlGetAttr(x, "ref")
   if(length(ref)) {
      doc = as(node, "XMLInternalDocument")
      node = getNodeSet(doc, "//x:*[@name =", sQuote(ref), " or @name =", sQuote(discardNamespace(ref)), "]", "x")
      if(length(node))
        return(xmlName(node[[1]]))
   }

   NA
}

getElementName =
  #
  # And defined differently again!!!
  #
function(node)
  xmlGetAttr(node, "name", xmlGetAttr(node, "ref", as.character(NA)))


asCount =
function(x)
{
   if(x == "unbounded")
     Inf
   else
     as.numeric(x)
}

getElementCount =
function(node)
{
  c(min = xmlGetAttr(node, "minOccurs", 1L, as.integer),
    max = xmlGetAttr(node, "maxOccurs", 1L, asCount))
}


processSequence =
  #
  # There are two basic kinds of sequences:
  #   1) an ordered collection of 1 or more different elements, some optional
  #   2) zero or more instances of the same element
  #
  #  2) maps to a list in R, but we might need to impose a constraint on the number of entries in the list.
  #  1) maps to a class definition. It is a structure, perhaps with missing/NULL/default values for  elements.
  #
  # There are also choice groups
  #
  #
function(node, types, namespaceDefs = list(), name = getElementName(node))
{
  if(xmlSize(node) == 1 && !is.na(xmlGetAttr(node[[1]], "maxOccurs", NA))) {

     ans = new("SimpleSequenceType", name = name,
                                      elType = SOAPType(xmlGetAttr(node[[1]], "type", xmlGetAttr(node[[1]], "ref", as.character(NA))),
                                                        nsuri = xmlNamespace(node),
                                                        namespaceDefs = namespaceDefs,
                                                        counts = getElementCount(node[[1]]), obj = new("SOAPTypeReference")))
     if(all(!is.na(ans@elType@count)) && all(ans@elType@count == 1)) {
          # if the sequence has a single element and the minOccurs and maxOccurs are both 1,
          # then return just the element.
        #XXX have to be careful that we recognize that the content is within the outer node given by name
       tmp = ans@elType
       return(tmp)
     }

     return(ans)
  
  }
  
  slotTypes = xmlApply(node, function(x) SOAPType(xmlGetAttr(x, "type", xmlGetAttr(x, "ref")), nsuri = xmlNamespace(x),
                                                  namespaceDefs = namespaceDefs, count = getElementCount(x)))
  names(slotTypes) = ids = as.character(xmlSApply(node, function(x) xmlGetAttr(x, "name", xmlGetAttr(x, "ref", NA))))

  slotTypes = slotTypes[ ! sapply(slotTypes, is.null) ]
  
  if(!is.na(name))
    ClassDef(name, slotTypes)
  else
    slotTypes
  
#  slotTypes[ !is.na(ids) ]
}


# Children
# no children - simple element  esearch.xsd  Count.
# complexType with a sequence  egquery.xsd  Result
# simpletype

setClass("SchemaCollection", contains = "list")
setClass("SchemaTypes", contains = "list")

setClassUnion("SOAPTypeOrNULL", c("SOAPType", "NULL"))

setClass("SchemaElement", representation(name = "character",  attributes = "list"), contains = "BasicSOAPType")  # added Sep 18 as an experiment.
setClass("SimpleElement", representation(type = "character"), contains = "SchemaElement")

setClass("Element", representation(type = "SOAPTypeOrNULL"),
                     contains = "SchemaElement")


setClassUnion("AttributeType", c("EnumValuesDef", "character"))

setClass("AttributeDef", representation(name = "character",
                                        type = "AttributeType",
                                        use = "character",  # prohibited, optional, required
                                        default = "character",
                                        fixed = "character"
                                        ),
                         prototype = list(use = "optional"))


processSchemaElement =
function(element, name = xmlGetAttr(element, "name"), namespaceDefs = list(), types = NULL)
{
#  return(element)

# if(xmlSize(element) > 0) {
#     warning("element with body")
#     return(element)
# }

  attrs = xmlAttrs(element)
  if(length(attrs) == 2 && all( names(attrs) %in% c("name", "type"))) 
    return(new("SimpleElement", name = attrs["name"], type = attrs["type"]))

  if(xmlSize(element) == 0) {
    obj =  new("SimpleElement", name = name)
  } else if(names(element)[1] == "complexType" && all(names(element[[1]]) == "attribute")) {
     # e.g. ParameterField, ArrayType in PMML.
    obj = new("SimpleElement", name = xmlGetAttr(element, "name", as.character(NA)),
                                attributes = xmlApply(element[[1]], processAttribute),
                                type = character(),
                                count = getElementCount(element))
    
  } else {
      # complexType
    if(xmlName(element[[1]]) == "complexType") {
        type = processSchemaType(element[[1]], types, namespaceDefs)
        obj = new("Element", name = name, type = type)

        i = xmlSApply(element[[1]], xmlName) == "attribute"
        if(any(i))  {
          obj@attributes = sapply(xmlChildren(element[[1]])[i], processAttribute)
          names(obj@attributes) = sapply(obj@attributes, slot, "name")
        }

        # obj@elements = xmlApply(element[[1]], processSchemaType, types, namespaceDefs)
        
      } else
         obj = NULL  
  } 

  obj
}


processAttributeGroup =
function(node)
{
  xmlApply(node, processAttribute)
}

processAttribute =
function(node, name = xmlGetAttr(node, "name"), type = xmlGetAttr(node, "type", as.character(NA)))
{
   if(is.na(type) && xmlSize(node)) {
     if(xmlName(node[[1]]) == "simpleType" &&
           xmlName(node[[1]][[1]]) == "restriction") {
        # handle non-string types too in the restriction.
       type = new("EnumValuesDef", values = xmlSApply(node[[1]][[1]], xmlGetAttr, "value"))
# Should be        
#       type = processSchemaType(node[[1]])
     } else
        warning("<fixme> Skipping children in <attribute> definition")
   }
   
   new("AttributeDef", name = name,
                       type = type,
                       use = xmlGetAttr(node, "use", "optional"),
                       default = xmlGetAttr(node, "default", as.character(NA)),       
                       fixed = xmlGetAttr(node, "fixed", as.character(NA)))
}



getSubstitutionGroups =
  #
  # doc = xmlParse("../inst/samples/kml21.xsd")
  # g = getSubstitutionGroups(doc)
  # by(g, g$group, function(x) x)
  #
  # Need to include this information in the class definitions.
  #
  #  e.g. kml:Geometry is a kml:Geometry and there are
  # the following in the substitutionGroup
  #         name                  type        group  
  #MultiGeometry kml:MultiGeometryType kml:Geometry
  #        Point         kml:PointType kml:Geometry
  #   LineString    kml:LineStringType kml:Geometry
  #   LinearRing    kml:LinearRingType kml:Geometry
  #      Polygon       kml:PolygonType kml:Geometry
  #        Model         kml:ModelType kml:Geometry  
  #
  # XXX Check for nesting in groups
  # 
  #
function(doc)
{
   groups = getNodeSet(doc, "//*[@substitutionGroup]")
   data.frame(name = sapply(groups, xmlGetAttr, "name"),
              type = sapply(groups, xmlGetAttr, "type"),
              group = sapply(groups, xmlGetAttr, "substitutionGroup"))
}
